@@ -23,7 +23,8 @@ from scipy.spatial import Delaunay
 from scipy.spatial.transform import Rotation as R
 
 
-from .Helpers import buildCameraMatrix, get_plane_infos, get_z_line_equation, get_plane_line_intersection
+from .helpers.geometry.euclidian import buildCameraMatrix, get_plane_infos, get_z_line_equation, get_plane_line_intersection
+from .helpers.geometry.orientation import rotateLandmarks
 
 # Get an instance of drawing specs to be used for drawing masks on faces
 DrawingSpec =  mp.solutions.drawing_utils.DrawingSpec
@@ -68,6 +69,10 @@ class Face():
                                     474, # Left
                                     475  # Top
                                 ]
+    left_eye_left_right_indices = [
+                                    362, # Right
+                                    263, # Left
+                                ]                                
     left_eye_center_index = 473
 
     simplified_right_eyelids_indices = [
@@ -103,6 +108,10 @@ class Face():
                                     469, # left
                                     470  # top
                                 ]
+    right_eye_left_right_indices = [
+                                    130, # Right
+                                    133, # Left
+                                ]                                   
     right_eye_center_index = 468    
 
     # Mouth
@@ -196,8 +205,98 @@ class Face():
                             4,  # top
                             45,
                             220,
-                            115
-                            ]  
+                            115,
+                            49,
+                            131,
+                            134,
+                            51,
+                            5,
+                            281,
+                            363,
+                            360,
+                            279,
+                            429,
+                            420,
+                            456,
+                            248,
+                            195,
+                            3,
+                            236,
+                            198,
+                            209,
+                            217,
+                            174,
+                            196,
+                            197,
+                            419,
+                            399,
+                            437,
+                            343,
+                            412,
+                            351,
+                            6,
+                            122,
+                            188,
+                            114,
+                            245,
+                            193,
+                            168,
+                            417,
+                            465
+                            ]
+
+    forehead_indices = [
+        301,
+        298,
+        333,
+        299,
+        337,
+        151,
+        108,
+        69,
+        104,
+        68,
+        71,
+        21,
+        54,
+        103,
+        67,
+        109,
+        10,
+        338,
+        297,
+        332,
+        284,
+        251
+    ]
+    # Eye_brows
+    left_eye_brows_indices = [
+        
+        285,
+        295,
+        282,
+        283,
+        276,
+        383,
+        300,
+        293,
+        334,
+        296,
+
+    ]
+    right_eye_brows_indices = [
+
+        156,
+        46,
+        53,
+        52,
+        65,
+        55,
+        66,
+        105,
+        63,
+        70
+    ]    
     # A list of simplified facial features used to reduce computation cost of drawing and morphing faces
     simplified_face_features = [
         10, 67, 54, 162, 127, 234, 93, 132,172,150,176,148,152,377,378,365,435,323,447,454,264,389,251, 332, 338, #Oval
@@ -208,7 +307,44 @@ class Face():
         50, 207, 280, 427
     ]
 
-
+    face_oval_indices = [
+        10,
+        109,
+        67,
+        103,
+        54,
+        21,
+        162,
+        127,
+        234,
+        93,
+        132,
+        58,
+        172,
+        136,
+        150,
+        149,
+        176,
+        148,
+        152,
+        377,
+        400,
+        378,
+        379,
+        365,
+        397,
+        288,
+        361,
+        323,
+        454,
+        356,
+        389,
+        251,
+        284,
+        332,
+        297,
+        338
+    ]
 
 
     all_face_features = list(range(468))
@@ -221,13 +357,14 @@ class Face():
         """
         self.image_shape = image_shape
 
-        self.update(landmarks)
+        if type(landmarks)==np.ndarray:
+            self.npLandmarks=landmarks
+        else:
+            self.update(landmarks)
 
 
         self.blinking = False
 
-        self.face_oval = list(set(
-            list(sum(list(mp.solutions.face_mesh.FACEMESH_FACE_OVAL), ()))))
 
         self.face_contours = list(set(
             list(sum(list(mp.solutions.face_mesh.FACEMESH_CONTOURS), ()))[::3]
@@ -401,6 +538,31 @@ class Face():
 
         return self.npLandmarks[indices,...]
 
+    def get_3d_realigned_landmarks_pos(self, indices: list=None,camera_matrix: np.ndarray = None, dist_coeffs: np.ndarray = np.zeros((4, 1))) -> np.ndarray:
+        """Returns a realigned version of the landmarks such that the head is looking forwards.
+        First the face orientation is computed, then the inverse rotation is applyed so that the face is facing the camera.
+        Useful for face recognition
+
+        Args:
+            indices (list): Indices of the landmarks to extract. Defaults to None, which means all landmarks
+
+        Returns:
+            np.ndarray: A realigned landmars vector of form nX3
+        """
+        # Correct orientation
+        vertices = self.npLandmarks.copy()
+        pos, ori = self.get_head_posture(camera_matrix, dist_coeffs)
+        if pos is not None :
+            center = vertices.mean(axis=0)
+            centered = (vertices-center)
+            vertices = rotateLandmarks(centered, ori, True)
+            vertices[:,1]*=-1
+            vertices += center
+        if indices is not None:
+            return vertices[indices,:]
+        else:
+            return vertices
+
     def get_realigned_landmarks_pos(self, indices: list=None) -> np.ndarray:
         """Returns a realigned version of the landmarks such that the head top is exactly in the center top and the chin is in the center  bottom
 
@@ -425,6 +587,29 @@ class Face():
         else:
             return vertices
 
+    def get_3d_realigned_face(self, camera_matrix: np.ndarray = None, dist_coeffs: np.ndarray = np.zeros((4, 1))):
+        """Returns a face object using a realigned version of the landmarks such that the head is looking forwards.
+        First the face orientation is computed, then the inverse rotation is applyed so that the face is facing the camera.
+        Useful for face recognition
+
+        Args:
+            indices (list): Indices of the landmarks to extract. Defaults to None, which means all landmarks
+
+        Returns:
+            np.ndarray: A realigned landmars vector of form nX3
+        """
+        # Correct orientation
+        vertices = self.npLandmarks.copy()
+        pos, ori = self.get_head_posture(camera_matrix, dist_coeffs)
+        if pos is not None :
+            center = vertices.mean(axis=0)
+            centered = (vertices-center)
+            vertices = rotateLandmarks(centered, ori, True)
+            vertices[:,1]*=-1
+            vertices += center
+
+        return Face(vertices,self.image_shape)
+
     def draw_landmark_by_index(self, image: np.ndarray, index: int, color: tuple = (255, 0, 0), radius: int = 5, thickness:int=1) -> np.ndarray:
         """Draw a landmark on an image from landmark index
 
@@ -485,7 +670,7 @@ class Face():
             image,(int(pos[0]), int(pos[1])), radius, color, thickness
         )
 
-    def draw_contour(self, image: np.ndarray, contour: np.ndarray, color: tuple = (255, 0, 0), thickness: int = 1) -> np.ndarray:
+    def draw_contour(self, image: np.ndarray, contour: np.ndarray, color: tuple = (255, 0, 0), thickness: int = 1, isClosed:bool = True) -> np.ndarray:
         """Draw a contour on an image
 
         Args:
@@ -493,6 +678,7 @@ class Face():
             contour (np.ndarray): a nX3 ndarray containing the positions of the landmarks
             color (tuple, optional): Color of the landmark. Defaults to (255, 0, 0).
             thickness (int, optional): Thickness of the line to draw the landmark. Defaults to 5.
+            isClosed (bool, optional): If True, the contour will be closed, otherwize it will be kept open. Defaults to True 
 
 
         Returns:
@@ -500,7 +686,7 @@ class Face():
         """
 
         pts = np.array([[int(p[0]), int(p[1])] for p in contour.tolist()]).reshape((-1, 1, 2))
-        return cv2.polylines(image, [pts], True, color, thickness)
+        return cv2.polylines(image, [pts], isClosed, color, thickness)
 
     def draw_overlay_on_left_iris(self, image:np.ndarray, overlay:np.ndarray)->np.ndarray:
         """Draws an overlay image on the left iris of the face
@@ -592,7 +778,7 @@ class Face():
 
         return face_pos, face_ori
 
-    def get_eyes_position(self)->tuple:
+    def get_eyes_position(self, camera_matrix:np.ndarray = None, dist_coeffs:np.ndarray=np.zeros((4,1)))->tuple:
         """Gets the posture of the eyes (position in cartesian space and Euler angles)
         Args:
             camera_matrix (int, optional)       : The camera matrix built using buildCameraMatrix Helper function. Defaults to a perfect camera matrix 
@@ -604,11 +790,14 @@ class Face():
         # Assertion to verify that the face object is ready
         assert self.ready, "Face object is not ready. There are no landmarks extracted."
 
+        # Correct orientation
+        lm = self.get_3d_realigned_landmarks_pos(camera_matrix=camera_matrix, dist_coeffs=dist_coeffs)
+
         # Left eye
-        iris = np.array(self.get_landmark_pos(Face.left_eye_center_index))
+        iris = np.array(lm[Face.left_eye_center_index,...])
         
-        left = np.array(self.get_landmark_pos(263))
-        right = np.array(self.get_landmark_pos(362))
+        left = np.array(lm[263])
+        right = np.array(lm[362])
 
         center = (left+right)/2
         ex = left-right
@@ -622,9 +811,9 @@ class Face():
         left_pos = np.array([np.dot((iris-center),ex)/nx,np.dot((iris-center),ey)/nx])
 
         # right
-        iris = np.array(self.get_landmark_pos(Face.right_eye_center_index))
-        left = np.array(self.get_landmark_pos(133))
-        right = np.array(self.get_landmark_pos(130))
+        iris = np.array(lm[Face.right_eye_center_index])
+        left = np.array(lm[133])
+        right = np.array(lm[130])
 
         center = (left+right)/2
         ex = left-right
@@ -638,7 +827,7 @@ class Face():
 
         return left_pos, right_pos
 
-    def compose_eye_rot(self, eye_pos:list, face_orientation:np.ndarray, offset=np.array([0,0]), x2ang: int=90, y2ang:int=60)->np.ndarray:
+    def compose_eye_rot(self, eye_pos:list, face_orientation:np.ndarray, offset=np.array([0,0]), x2ang: int=180, y2ang:int=60)->np.ndarray:
         """Composes eye position with face rotation to produce eye orientation in world coordinates
 
         Args:
@@ -802,8 +991,6 @@ class Face():
         right_eye_iris_lower = right_eye_contour[3, ...]
 
 
-
-
         if draw_landmarks:
 
             image = self.draw_landmark(image, left_eye_upper, (0, 0, 255),1)
@@ -927,8 +1114,11 @@ class Face():
             landmarks = self.npLandmarks[:, :2]
         else:
             landmarks = self.npLandmarks[landmark_indices, :2]
-
-        triangleList = self.triangles
+        try:
+            triangleList = self.triangles
+        except:
+            self.triangulate(landmark_indices)
+            triangleList = self.triangles
         size = img.shape
         r = (0, 0, size[1], size[0])
         ncolors = len(delaunay_colors)
@@ -947,7 +1137,7 @@ class Face():
 
         return img
 
-    def getFaceBox(self, image:np.ndarray, landmark_indices:list=None)->np.ndarray:
+    def getFaceBox(self, image:np.ndarray, landmark_indices:list=None, margins=(0,0,0,0))->np.ndarray:
         """Gets an image of the face extracted from the original image (simple box extraction which will extract some of the background)
 
         Args:
@@ -966,8 +1156,8 @@ class Face():
             landmarks = self.npLandmarks[:, :2]
         else:
             landmarks = self.npLandmarks[landmark_indices, :2]
-        p1 = landmarks.min(axis=0)
-        p2 = landmarks.max(axis=0)
+        p1 = landmarks.min(axis=0)-np.array(margins[0:2])
+        p2 = landmarks.max(axis=0)+np.array(margins[2:4])
         return image[int(p1[1]):int(p2[1]),int(p1[0]):int(p2[0])]
 
     def getFace(self, image:np.ndarray, src_triangles:list(), landmark_indices:list=None)->np.ndarray:
@@ -999,25 +1189,37 @@ class Face():
         dest = np.zeros_like(croped)
         for tr in src_triangles:
             mask = np.zeros_like(dest)
-            n_mask = np.ones_like(dest)
             try:
                 t_dest = np.array([landmarks[tr[0], 0:2], landmarks[tr[1], 0:2], landmarks[tr[2], 0:2]])
                 cv2.fillConvexPoly(mask, t_dest.astype(np.int), [1, 1, 1])
-                cv2.fillConvexPoly(n_mask, t_dest.astype(np.int), [0, 0, 0])
-                dest = cv2.multiply(dest, n_mask) + cv2.multiply(mask, croped)
+                mask=mask.astype(np.bool)
+                dest = dest*~mask + mask* croped
             except Exception as ex:
                 pass
         return dest
 
-    def copyToFace(self, dst_face, src_image, dst_image:np.ndarray, landmark_indices:list=None, opacity:float=1.0)->np.ndarray:
+    def copyToFace(
+                    self, 
+                    dst_face, 
+                    src_image:np.ndarray, 
+                    dst_image:np.ndarray, 
+                    landmark_indices:list=None, 
+                    opacity:float=1.0, 
+                    min_input_triangle_cross:float=20, 
+                    min_output_triangle_cross:float=1, 
+                    retriangulate:bool=False,
+                    seemless_cloning:bool=False,
+                    empty_fill_color:Tuple=None
+                    )->np.ndarray:
         """Copies the face to another image (used for face copy or face morphing)
 
         Args:
             dst_face (Face): A face object describing the face in the destination image (triangulate function should have been called on this object with the same landmark_indices argument)
-            src_image ([type]): [description]
+            src_image (np.ndarray): [description]
             dst_image (np.ndarray): [description]
             landmark_indices (list, optional): The list of landmarks to be used (the same list used for the triangulate method that allowed the extraction of the triangles). Defaults to None.
             opacity (int, optional): the opacity level of the face (between 0 and 1)
+            retriangulate (bool): if true, then triangles will be computed avery time (needed if the source face is moving)
 
         Returns:
             np.ndarray: An image containing only the face
@@ -1034,55 +1236,143 @@ class Face():
             src_landmarks = self.npLandmarks[landmark_indices, :2]
             dst_landmarks = dst_face.npLandmarks[landmark_indices, :2]
 
+        # Clip landmarks to image size
+        src_landmarks = np.clip(src_landmarks,np.array([0,0]),np.array([src_image.shape[1],src_image.shape[0]])).astype(np.int)
+        dst_landmarks = np.clip(dst_landmarks,np.array([0,0]),np.array([src_image.shape[1],src_image.shape[0]])).astype(np.int)
+
+
         # Crop images
         src_p1 = src_landmarks.min(axis=0)
         src_p2 = src_landmarks.max(axis=0)
 
         src_landmarks -= src_p1
-        src_crop = src_image[int(src_p1[1]):int(
-            src_p2[1]), int(src_p1[0]):int(src_p2[0])]
+        src_crop = src_image[src_p1[1]:src_p2[1], src_p1[0]:src_p2[0]]
 
         dst_p1 = dst_landmarks.min(axis=0)
         dst_p2 = dst_landmarks.max(axis=0)
 
         dst_landmarks -= dst_p1
-        dst_crop = dst_image[int(dst_p1[1]):int(
-            dst_p2[1]), int(dst_p1[0]):int(dst_p2[0])]
+        dst_crop = dst_image[dst_p1[1]:dst_p2[1], dst_p1[0]:dst_p2[0]]
 
         # Prepare empty image
-        dest = np.zeros((int(dst_p2[1])-int(dst_p1[1]),int(dst_p2[0])-int(dst_p1[0]),3))
+        dest = dst_crop.copy()# np.zeros_like(dst_crop)
+
+
         dst_h, dst_w, _ = dest.shape
         center = (dst_w//2,dst_h//2)
 
+        if retriangulate:
+            dst_face.triangulate(landmark_indices)
+
+        final_mask=np.zeros_like(dst_crop,dtype=np.uint8)
+        # Prepare masks
         for src_tr in self.triangles:
-            mask = np.zeros_like(dest)
-            n_mask = np.ones_like(dest)
-            try:
+            #try:
+                fill=False
+                # Get source triangles
                 t_src = np.array(
-                    [src_landmarks[src_tr[0], 0:2], src_landmarks[src_tr[1], 0:2], src_landmarks[src_tr[2], 0:2]])
+                    [
+                        src_landmarks[src_tr[0], 0:2], 
+                        src_landmarks[src_tr[1], 0:2], 
+                        src_landmarks[src_tr[2], 0:2]
+                    ])
+                # Get destination triangles
                 t_dest = np.array(
-                    [dst_landmarks[src_tr[0], 0:2], dst_landmarks[src_tr[1], 0:2], dst_landmarks[src_tr[2], 0:2]])
+                    [
+                        dst_landmarks[src_tr[0], 0:2], 
+                        dst_landmarks[src_tr[1], 0:2], 
+                        dst_landmarks[src_tr[2], 0:2]
+                    ])
+                
+                # Test triangles are not empty
+                v1 = t_src[1,:]-t_src[0,:]
+                v2 = t_src[2,:]-t_src[1,:]
+                cross = np.abs(np.cross(v1,v2))
+                if cross<min_input_triangle_cross:
+                    if empty_fill_color is not None:
+                        fill=True
+                    else:
+                        continue
+                v1 = t_dest[1,:]-t_dest[0,:]
+                v2 = t_dest[2,:]-t_dest[1,:]
+                cross = np.abs(np.cross(v1,v2))
+                if cross<min_output_triangle_cross:
+                    continue
+                
 
-                cv2.fillConvexPoly(mask, t_dest.astype(np.int), [1, 1, 1])
-                cv2.fillConvexPoly(n_mask, t_dest.astype(np.int), [0, 0, 0])
+                #Crop more to just get the triangle zone
+                """
+                """
+                src_min_pos = np.clip(t_src.min(axis=0)-np.array([2,2]),np.array([0,0]),np.array([src_image.shape[1],src_image.shape[0]]))
+                src_max_pos = np.clip(t_src.max(axis=0)+np.array([2,2]),np.array([0,0]),np.array([src_image.shape[1],src_image.shape[0]]))
 
+                dst_min_pos = np.clip(t_dest.min(axis=0)-np.array([2,2]),np.array([0,0]),np.array([dest.shape[1],dest.shape[0]]))
+                dst_max_pos = np.clip(t_dest.max(axis=0)+np.array([2,2]),np.array([0,0]),np.array([dest.shape[1],dest.shape[0]]))
+                
+                t_src -= src_min_pos
+                t_dest -= dst_min_pos
+
+                #Find transformation matrix from source triangle to destination triangle
                 M = cv2.getAffineTransform(
                     t_src.astype(np.float32),
                     t_dest.astype(np.float32)
                 )
-                warped = cv2.warpAffine(src_crop,  # src_image,
-                                        M,
-                                        (dst_w, dst_h)
-                                        )
-                dest = cv2.multiply(dest, n_mask) + cv2.multiply(np.float64(warped), mask)
-            except Exception as ex:
-                pass
-        mask = cv2.threshold(dest.astype(np.uint8), 1, opacity*255, cv2.THRESH_BINARY)[1]
-        try:
-            dst_crop = cv2.seamlessClone(dest.astype(np.uint8), dst_crop, mask, (int(center[0]),int(center[1])), cv2.NORMAL_CLONE)
-        except:
-            pass
-        dst_image[int(dst_p1[1]):int(dst_p2[1]), int(dst_p1[0]):int(dst_p2[0])] = dst_crop
+
+                src_w,src_h = src_max_pos[0]-src_min_pos[0], src_max_pos[1]-src_min_pos[1]
+                dest_w,dest_h = dst_max_pos[0]-dst_min_pos[0], dst_max_pos[1]-dst_min_pos[1]
+
+                # If the cropped size is zero, go to next triangle
+                if src_w<=1 or src_h<=1 or dest_w<=1 or dest_h<=1:
+                    continue
+
+                if fill:
+                    stem = np.ones((src_max_pos[1]-src_min_pos[1],
+                                    src_max_pos[0]-src_min_pos[0]))
+                    warped_t = cv2.warpAffine(np.dstack(
+                                                    [
+                                                        (stem*empty_fill_color[0])[:,:,None],
+                                                        (stem*empty_fill_color[1])[:,:,None],
+                                                        (stem*empty_fill_color[2])[:,:,None]
+                                                    ]),
+                                            M,
+                                            (   dest_w,dest_h
+                                            ),flags=cv2.INTER_LINEAR 
+                                            )
+                else:
+                    warped_t = cv2.warpAffine(src_crop[ src_min_pos[1]:src_max_pos[1],
+                                                        src_min_pos[0]:src_max_pos[0]], 
+                                            M,
+                                            (   dest_w,dest_h
+                                            ),flags=cv2.INTER_LINEAR
+                                            )
+                #Build masks
+                mask = np.zeros_like(warped_t)
+                # Prepare masks
+                cv2.fillConvexPoly(mask, t_dest, [1, 1, 1])
+                mask = mask.astype(np.bool)
+
+                # Build global mask
+                final_mask[dst_min_pos[1]:dst_max_pos[1],dst_min_pos[0]:dst_max_pos[0],:] = final_mask[dst_min_pos[1]:dst_max_pos[1],dst_min_pos[0]:dst_max_pos[0],:]*~mask+ mask*255
+
+                # Copy warped triangle to destination
+                warped_masked = warped_t* mask # cv2.dilate(warped_t* mask,np.ones((2,2)))* mask
+                dest_masked = dest[dst_min_pos[1]:dst_max_pos[1],dst_min_pos[0]:dst_max_pos[0],:]* ~mask#cv2.dilate(dest[dst_min_pos[1]:dst_max_pos[1],dst_min_pos[0]:dst_max_pos[0],:]* ~mask,np.ones((2,2)))* ~mask
+                dest[dst_min_pos[1]:dst_max_pos[1],dst_min_pos[0]:dst_max_pos[0],:] =  dest_masked + warped_masked
+
+            #except Exception as ex:
+            #    pass
+
+        if seemless_cloning:
+            dst_crop = cv2.seamlessClone(dest, dst_crop, final_mask, (int(center[0]),int(center[1])), cv2.NORMAL_CLONE)
+        else:
+            final_mask= final_mask.astype(np.bool)
+            dst_crop = dest*final_mask + dst_crop*~final_mask
+        #dst_crop = final_mask 
+        dst_image[
+            int(dst_p1[1]):int(dst_p2[1]),
+            int(dst_p1[0]):int(dst_p2[0]),
+            :
+            ] = dst_crop
         return dst_image
 
     def draw_bounding_box(self, image:np.ndarray, color:tuple=(255,0,0), thickness:int=1, text=None):
